{-# LANGUAGE OverloadedStrings, TupleSections #-}
import Text.Regex.Applicative
import Control.Monad.Trans.Reader
import System.Directory.Extra
import System.FilePath
import qualified Data.Map as M
import System.IO
import Data.List
import Data.Char
import Data.Maybe
import Control.Arrow

-- Specification

spec = do
    inFile "Arity" $ do
        extract "def-arity"
            (before (def "typedef" "Arity")) (before blankline)
        extract "def-arity-order"
            (before (def "instantiation" "Arity :: po")) (before blankline)
        extract "def-arity-zero"
            (before (def "instantiation" "Arity :: zero")) (before blankline)
        extract "def-arity-inc-pre"
            (before (def "lift_definition" "inc_Arity")) (before blankline)
        extract "def-arity-inc"
            (before (def "definition" "inc")) (before2nd blankline)
    inFile "TransformTools" $ do
        extract "def-lift-transform"
            (before (def "lemma" "lift_transform_simps")) (before proofdelim)
        extract "def-map-transform"
            (before (def "definition" "map_transform")) (before blankline)
    inFile "ArityTransform" $ do
        extract "def-transform"
            (before (def "lemma" "transform_simps")) (before proofdelim)
    inFile "CoCallGraph" $ do
        extract "def-cocalls"
            (before (def "typedef" "CoCalls")) (before blankline)
        extract "def-ccfield"
            (before (def "lift_definition" "ccField")) (before blankline)
        extract "def-incc"
            (before (def "lift_definition" "inCC")) (before2nd blankline)
        extract "def-ccdelete"
            (before (def "lift_definition" "cc_delete")) (before proofdelim)
        extract "def-ccrestr"
            (before (def "lift_definition" "cc_restr")) (before proofdelim)
        extract "def-ccprod"
            (before (def "lift_definition" "ccProd")) (before proofend)
        extract "def-ccsquare"
            (before (def "definition" "ccSquare")) (before blankline)
        extract "def-ccneighbors"
            (before (def "lift_definition" "ccNeighbors")) (before blankline)
    inFile "CoCallAnalysisImpl" $ do
        extract "def-aexp"
            (before (def "lemma" "Aexp_simps")) (before proofdelim)
        extract "def-ccexp"
            (before (def "lemma" "CCexp_simps")) (before proofdelim)
    inFile "ArityAnalysisSig" $ do
        extract "def-aexp-bot"
            (before (def "abbreviation" "Aexp_bot_syn")) (before2nd endline)
    inFile "CoCallAnalysisSig" $ do
        extract "def-ccexp-bot"
            (before (def "abbreviation" "ccExp_bot_syn")) (before2nd endline)
    inFile "ArityAnalysisAbinds" $ do
        extract "def-abind"
            (before (def "lemma" "ABind_eq")) (before proofdelim)
        extract "def-abinds"
            (before (def "fun" "ABinds")) (before blankline)
    inFile "CoCallAnalysisBinds" $ do
        extract "def-ccbind"
            (before (def "lemma" "ccBind_eq")) (before proofdelim)
        extract "def-ccbinds"
            (before (def "lemma" "ccBinds_eq")) (before proofdelim)
        extract "def-ccbindsextra"
            (before (def "lemma" "ccBindsExtra_eq")) (before proofdelim)
    inFile "CoCallFix" $ do
        extract "def-abind-nonrec"
            (before (def "lemma" "ABind_nonrec_eq")) (before proofdelim)
        extract "def-ccfix"
            (before (def "lemma" "CCfix_eq")) (before proofdelim)
    inFile "ArityAnalysisFix" $ do
        extract "def-afix"
            (before (def "lemma" "Afix_eq")) (before proofdelim)
    inFile "CoCallAnalysisImpl" $ do
        extract "def-aexp"
            (before (def "lemma" "Aexp_simps")) (before proofdelim)
        extract "def-ccexp"
            (before (def "lemma" "CCexp_simps")) (before proofdelim)
    inFile "CallArityEnd2EndSafe" $ do
        extract "call-artiy-safe-end2end"
            (before (def "theorem" "end2end_closed")) (before proofdelim)
    inFile "ArityAnalysisCorrDenotational" $ do
        extract "arity-transformation-correct"
            (before (def "corollary" "Arity_transformation_correct'")) (before proofdelim)
    inFile "Vars" $ do
        extract "def-var"
            (before (cmd "atom_decl")) (before endline)
    inFile "AList-Utils" $ do
        extract "def-domA"
            (before (def "definition" "domA")) (before blankline)
    inFile "Terms" $ do
        extract "exp-definition"
            (before (cmd "nominal_datatype")) (before blankline)
        extract "def-isval"
            (before (def "nominal_function" "isVal")) (before proofdelim)
        extract "def-heap"
            (before (def "type_synonym" "heap")) (before endline)
        extract "def-heapToAssn"
            (before (def "fun" "heapToAssn")) (before blankline)
        extract "def-Let"
            (before (def "definition" "Let")) (before blankline)
    inFile "Substitution" $ do
        extract "def-subst-var"
            (before (def "fun" "subst_var")) (before blankline)
        extract "def-subst"
            (before (cmd "nominal_function")) (before proofdelim)
    inFile "Launchbury" $ do
        extract "def-launchbury"
            (before (def "inductive" "reds")) (before blankline)
    inFile "SestoftConf" $ do
        extract "def-conf"
            (before (def "type_synonym" "conf")) (before blankline)
    inFile "Sestoft" $ do
        extract "def-step"
            (before (def "inductive" "step")) (before2nd blankline)
    inFile "Nominal-Utils" $ do
        extract "def-fv"
            (before (def "definition" "fv")) (before blankline)
    inFile "Env" $ do
        extract "def-edom"
            (before (def "definition" "edom")) (before blankline)
        extract "def-esing"
            (before (def "lemma" "esing_simps")) (before proofdelim)
        extract "def-env-delete"
            (before (def "definition" "env_delete")) (before blankline)
        extract "def-env-restr"
            (before (def "definition" "env_restr")) (skip (def "abbreviation" "env_restr_rev") (before blankline))
    inFile "Adequacy" $ do
        extract "thm-adequacy"
            (before (def "theorem" "adequacy")) (before proofdelim)
    inFile "CorrectnessOriginal" $ do
        extract "correctness-case-app"
            (after (() <$ string "nice proof start")) (before (() <$ string "nice proof end"))
        extract "thm-correctness"
            (before (def "theorem" "correctness")) (before proofdelim)
    inFile "AbstractDenotational" $ do
        extract "abstract-denotational-locale"
            (before (def "locale" "semantic_domain")) (before begin)
        extract "abstract-denotational-definition"
            (before (cmd "nominal_function")) (before proofdelim)
    inFile "EvalHeap" $ do
        extract "def-evalheap"
            (before (def "fun" "evalHeap")) (before blankline)
    inFile "HeapSemantics" $ do
        extract "def-hsem"
            (before (def "abbreviation" "HSem_syn")) (skip blankline (before proofdelim))
    inFile "HasESem" $ do
        extract "has-ESem-locale"
            (before (cmd "locale")) (before begin)
    inFile "HeapSemantics" $ do
        extract "HSem-definition"
            (before (def "definition" "HSem")) (before blankline)
    inFile "Denotational" $ do
        extract "def-esem-syntax"
            (before (def "abbreviation" "ESem_syn''")) (before blankline)
        extract "denotational-interpretation"
            (before (cmd "interpretation")) (before blankline)
        extract "denotational-equations"
            (before (def "lemma" "ESem_simps:")) (after proofend)
        extract "def-esem"
            (before (def "lemma" "ESem_simps:")) (before proofdelim)
    inFile "ResourcedDenotational" $ do
        extract "resourced-denotational-interpretation"
            (before (cmd "interpretation")) (before blankline)
    inFile "Value" $ do
        extract "def-value"
            (before (cmd "domain")) (skip blankline $ skip blankline $ before blankline)
        extract "value-domain"
            (before (cmd "domain")) (before blankline)
    inFile "CValue" $ do
        extract "cvalue-domain"
            (before (cmd "domain")) (before blankline)
    inFile "TTree" $ do
        extract "downset"
            (before (def "definition" "downset")) (before blankline)
        extract "ttree"
            (before (def "typedef" "'a ttree")) (before blankline)
        extract "paths-definition"
            (before (def "lift_definition" "paths")) (before blankline)
        extract "automata-view"
            (before (def "lift_definition" "possible")) (before2nd proofend)
        extract "either-definition"
            (before (def "lift_definition" "either")) (before proofend)
        extract "both-definition"
            (before (def "lift_definition" "both")) (before proofend)
        extract "without-definition"
            (before (def "lift_definition" "without")) (before proofend)
        extract "substitute-prime-definition"
            (before (def "definition" "f_nxt")) (before2nd blankline)
        extract "substitute-definition"
            (before (def "lift_definition" "substitute")) (before proofend)
        extract "substitute-prime-prime"
            (before (def "inductive" "substitute''")) (before proofdelim)
        extract "ttree-rest-substitute2"
            (before (def "lemma" "ttree_rest_substitute2")) (before proofdelim)
    inFile "ITree" $ do
        extract "ttree-holcf"
            (before (cmd "domain")) (before blankline)
    inFile "LTree" $ do
        extract "ttree-codatatype"
            (before (cmd "codatatype")) (before blankline)


-- Code

type R = RE Char ()
type Anchor = (R,R)

(<>) :: RE s a -> RE s b -> RE s ()
r1 <> r2 = () <$ r1 <* r2

cmd :: String -> R
cmd s = "\\isacommand{" <> isastring s <> "}"

def :: String -> String -> R
def c n = cmd c <> garbage <> isastring n
 where
    garbage = few $
        ws <|>
        () <$ "\\isamarkupfalse" <|>
        () <$ "\\isanewline" <|>
        () <$ "\\ " <|>
        (isastring "'" <> many (psym isAlpha))

before :: R -> Anchor
before r = (pure (),r)

before2nd:: R -> Anchor
before2nd r = skip r (before r)

skip :: R -> Anchor -> Anchor
skip r (b,a) = (r <> few anySym <> b, a)

after :: R -> Anchor
after r = (r, pure ())

blankline :: R
blankline = "\\isanewline" <> ws <> "\\isanewline"

endline :: R
endline = () <$ "\\isanewline"

proofdelim :: R
proofdelim = "\\isanewline" <> ws <> "\\isadelimproof"

begin :: R
begin = "\\isanewline" <> ws <> "\\isakeyword{begin}"

proofend :: R
proofend = () <$ "\\endisatagproof"

ws :: R
ws = () <$ many (psym (`elem` (" %\n"::[Char])))

isastring :: String -> R
isastring = (()<$) . string . concatMap go
  where go ' '  = "\\ "
        go '\'' = "{\\isacharprime}"
        go ':'  = "{\\isacharcolon}"
        go '_'  = "{\\isacharunderscore}"
        go c    | isDigit c
                = "{\\isadigit{" ++ [c] ++ "}}"
        go c    = [c]

matched :: RE s a -> RE s [s]
matched r = snd <$> withMatched r

extract :: String -> Anchor -> Anchor -> ReaderT Files IO ()
extract name (from0,from1) (to0,to1) = ReaderT $ \input -> do
    let re = from0 *> matched (from1 <> few anySym <> to0) <* to1
    case [ (fn,m) | (fn,i) <- input, Just m <- return $ findFirstInfix re i ] of
        [] -> hPutStrLn stderr $ "Could not find data for " ++ name ++ " (" ++ show (length input) ++ " files searched)"
        (ifn,(_,m,_)):_ -> do
            hPutStrLn stderr $ "Writing " ++ fn ++ " (" ++ show (length (lines m)) ++ " lines)"
            writeFile fn $
              "\\begin{isablocksource}{" ++ ifn ++ "}" ++ "\n" ++
              m ++ "\n" ++
              "\\end{isablocksource}"

  where fn = filename name

filename :: String -> FilePath
filename name = "isa-thy-output" </> name <.> "tex"

type Files = [(String, String)]

inFile :: String -> ReaderT Files m a -> ReaderT Files m a
inFile f = withReaderT (filter ((==f) . takeBaseName . fst))

main = do
    inputFiles <- filter (".tex" `isSuffixOf`) <$> listFiles "isa-thy-output/empty/"
    hPutStrLn stderr $ "Reading " ++ show (length inputFiles) ++ " files."
    input <- mapM (\n -> (takeBaseName n,) <$> readFile n) inputFiles
    hPutStrLn stderr $ "Done."

    {-
    thyFiles <- filter (".thy" `isSuffixOf`) . concat <$> mapM listFiles
        [ "/home/jojo/uni/info/isa-launchbury/Call_Arity/"
        , "/home/jojo/uni/info/isa-launchbury/Launchbury/"
        , "/home/jojo/uni/info/isa-launchbury/Scratchpad/"
        ]
    hPutStrLn stderr $ "Reading " ++ show (length thyFiles) ++ " theory files."
    theories <- M.fromList <$> mapM (\n -> (takeBaseName n,) <$> readFile n) inputFiles
    hPutStrLn stderr $ "Done."
    -}

    flip runReaderT input spec
